import json
from logging import Logger

from msgbus import TK_SYS_ACK
from posot import OrderTaker

from interactor.route.loyalty.search_customer_info.dto import SearchCustomerInfoResponse
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import (
    Message,
    DataType,
)
from model.error_type import NotFoundErrorsType
from model.exception import NotFoundError
from model.loyalty import CustomerInfo, Benefit, BenefitInfo

from model.services import LoyaltyServiceModel
from model.config import Config
from model.tokens import BusTokens
from typing import Dict


class LoyaltyService(LoyaltyServiceModel):

    def __init__(
        self,
        message_bus,    # type: MbContextMessageBus
        logger,         # type: Logger
        config,         # type: Config
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.logger = logger
        self.config = config

    def get_customer_info(
        self,
        pos_ot,         # type: OrderTaker
        customer_id,    # type: str
    ):
        # type: (...) -> SearchCustomerInfoResponse

        message = Message(
            token=BusTokens.TK_LOYALTY_GET_LOYALTY_ID,
            data_type=DataType.param,
            data=customer_id,
        )

        response = self.message_bus.send_message(
           component="Loyalty",
           message=message,
        )

        if response.token != TK_SYS_ACK or not response.data:
            self.logger.info("CustomerId: {}, not found".format(customer_id))
            raise NotFoundError(
                ex=NotFoundErrorsType.CustomerNotFound,
            )

        return self._process_response(
            pos_ot=pos_ot,
            response=response,
        )

    def _process_response(
        self,
        pos_ot,     # type: OrderTaker
        response,   # type: Message
    ):
        # type: (...) -> SearchCustomerInfoResponse

        try:
            loyalty = json.loads(
                response.data,
            )

            if loyalty in [None, ""]:
                self.logger.exception("Empty customer info")
                raise NotFoundError(
                    ex=NotFoundErrorsType.CustomerNotFound,
                )

            self._save_custom_properties(
                pos_ot=pos_ot,
                loyalty=loyalty,
            )

            return self._get_response(
                loyalty=loyalty,
            )
        except NotFoundError as ex:
            raise ex
        except Exception as _:
            self.logger.exception("Error process response loyalty customer info")
            raise NotFoundError(
                ex=NotFoundErrorsType.CustomerNotFound,
            )

    @staticmethod
    def _save_custom_properties(
        pos_ot,       # type: OrderTaker
        loyalty,      # type: Dict
    ):
        properties = {"LOYALTY_CUSTOMER_INFO": loyalty}

        customer_info = loyalty.get("customerInfo")

        if customer_info is not None:
            customer_data = customer_info.get("customer")

            if customer_data is not None:
                customer_document = customer_data.get("document", None)
                properties.update({"CUSTOMER_DOC": customer_document})

                customer_name = customer_data.get("name", None)
                properties.update({"CUSTOMER_NAME": customer_name})

            for prop in properties.keys():
                if properties[prop] in [None, ""]:
                    del properties[prop]

        pos_ot.setOrderCustomProperties(
           properties=properties,
        )

    @staticmethod
    def _get_response(
        loyalty,      # type: Dict
    ):
        # type: (...) -> SearchCustomerInfoResponse

        customer_info = CustomerInfo(
            loyalty=loyalty,
        )

        benefits = loyalty.get("benefits", [])
        list_benefit = []
        for benefit in benefits:
            current_benefit = Benefit(
               loyalty=benefit,
            )

            notify = benefit.get("notify", False)
            benefit_info = BenefitInfo(
                benefit=current_benefit,
                notify=notify,
            )
            list_benefit.append(benefit_info)

        response = SearchCustomerInfoResponse(
            customer_info=customer_info,
            benefits=list_benefit,
        )

        return response



# -*- coding: utf-8 -*-
"""
    werkzeug.debug.tbtools
    ~~~~~~~~~~~~~~~~~~~~~~

    This module provides various traceback related utility functions.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import codecs
import inspect
import json
import os
import re
import sys
import sysconfig
import traceback
from tokenize import TokenError

from .._compat import PY2
from .._compat import range_type
from .._compat import reraise
from .._compat import string_types
from .._compat import text_type
from .._compat import to_native
from .._compat import to_unicode
from ..filesystem import get_filesystem_encoding
from ..utils import cached_property
from ..utils import escape
from .console import Console


_coding_re = re.compile(br"coding[:=]\s*([-\w.]+)")
_line_re = re.compile(br"^(.*?)$", re.MULTILINE)
_funcdef_re = re.compile(r"^(\s*def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)")
UTF8_COOKIE = b"\xef\xbb\xbf"

system_exceptions = (SystemExit, KeyboardInterrupt)
try:
    system_exceptions += (GeneratorExit,)
except NameError:
    pass


HEADER = u"""\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
  "http://www.w3.org/TR/html4/loose.dtd">
<html>
  <head>
    <title>%(title)s // Werkzeug Debugger</title>
    <link rel="stylesheet" href="?__debugger__=yes&amp;cmd=resource&amp;f=style.css"
        type="text/css">
    <!-- We need to make sure this has a favicon so that the debugger does
         not by accident trigger a request to /favicon.ico which might
         change the application state. -->
    <link rel="shortcut icon"
        href="?__debugger__=yes&amp;cmd=resource&amp;f=console.png">
    <script src="?__debugger__=yes&amp;cmd=resource&amp;f=jquery.js"></script>
    <script src="?__debugger__=yes&amp;cmd=resource&amp;f=debugger.js"></script>
    <script type="text/javascript">
      var TRACEBACK = %(traceback_id)d,
          CONSOLE_MODE = %(console)s,
          EVALEX = %(evalex)s,
          EVALEX_TRUSTED = %(evalex_trusted)s,
          SECRET = "%(secret)s";
    </script>
  </head>
  <body style="background-color: #fff">
    <div class="debugger">
"""
FOOTER = u"""\
      <div class="footer">
        Brought to you by <strong class="arthur">DON'T PANIC</strong>, your
        friendly Werkzeug powered traceback interpreter.
      </div>
    </div>

    <div class="pin-prompt">
      <div class="inner">
        <h3>Console Locked</h3>
        <p>
          The console is locked and needs to be unlocked by entering the PIN.
          You can find the PIN printed out on the standard output of your
          shell that runs the server.
        <form>
          <p>PIN:
            <input type=text name=pin size=14>
            <input type=submit name=btn value="Confirm Pin">
        </form>
      </div>
    </div>
  </body>
</html>
"""

PAGE_HTML = (
    HEADER
    + u"""\
<h1>%(exception_type)s</h1>
<div class="detail">
  <p class="errormsg">%(exception)s</p>
</div>
<h2 class="traceback">Traceback <em>(most recent call last)</em></h2>
%(summary)s
<div class="plain">
  <form action="/?__debugger__=yes&amp;cmd=paste" method="post">
    <p>
      <input type="hidden" name="language" value="pytb">
      This is the Copy/Paste friendly version of the traceback.  <span
      class="pastemessage">You can also paste this traceback into
      a <a href="https://gist.github.com/">gist</a>:
      <input type="submit" value="create paste"></span>
    </p>
    <textarea cols="50" rows="10" name="code" readonly>%(plaintext)s</textarea>
  </form>
</div>
<div class="explanation">
  The debugger caught an exception in your WSGI application.  You can now
  look at the traceback which led to the error.  <span class="nojavascript">
  If you enable JavaScript you can also use additional features such as code
  execution (if the evalex feature is enabled), automatic pasting of the
  exceptions and much more.</span>
</div>
"""
    + FOOTER
    + """
<!--

%(plaintext_cs)s

-->
"""
)

CONSOLE_HTML = (
    HEADER
    + u"""\
<h1>Interactive Console</h1>
<div class="explanation">
In this console you can execute Python expressions in the context of the
application.  The initial namespace was created by the debugger automatically.
</div>
<div class="console"><div class="inner">The Console requires JavaScript.</div></div>
"""
    + FOOTER
)

SUMMARY_HTML = u"""\
<div class="%(classes)s">
  %(title)s
  <ul>%(frames)s</ul>
  %(description)s
</div>
"""

FRAME_HTML = u"""\
<div class="frame" id="frame-%(id)d">
  <h4>File <cite class="filename">"%(filename)s"</cite>,
      line <em class="line">%(lineno)s</em>,
      in <code class="function">%(function_name)s</code></h4>
  <div class="source %(library)s">%(lines)s</div>
</div>
"""

SOURCE_LINE_HTML = u"""\
<tr class="%(classes)s">
  <td class=lineno>%(lineno)s</td>
  <td>%(code)s</td>
</tr>
"""


def render_console_html(secret, evalex_trusted=True):
    return CONSOLE_HTML % {
        "evalex": "true",
        "evalex_trusted": "true" if evalex_trusted else "false",
        "console": "true",
        "title": "Console",
        "secret": secret,
        "traceback_id": -1,
    }


def get_current_traceback(
    ignore_system_exceptions=False, show_hidden_frames=False, skip=0
):
    """Get the current exception info as `Traceback` object.  Per default
    calling this method will reraise system exceptions such as generator exit,
    system exit or others.  This behavior can be disabled by passing `False`
    to the function as first parameter.
    """
    exc_type, exc_value, tb = sys.exc_info()
    if ignore_system_exceptions and exc_type in system_exceptions:
        reraise(exc_type, exc_value, tb)
    for _ in range_type(skip):
        if tb.tb_next is None:
            break
        tb = tb.tb_next
    tb = Traceback(exc_type, exc_value, tb)
    if not show_hidden_frames:
        tb.filter_hidden_frames()
    return tb


class Line(object):
    """Helper for the source renderer."""

    __slots__ = ("lineno", "code", "in_frame", "current")

    def __init__(self, lineno, code):
        self.lineno = lineno
        self.code = code
        self.in_frame = False
        self.current = False

    @property
    def classes(self):
        rv = ["line"]
        if self.in_frame:
            rv.append("in-frame")
        if self.current:
            rv.append("current")
        return rv

    def render(self):
        return SOURCE_LINE_HTML % {
            "classes": u" ".join(self.classes),
            "lineno": self.lineno,
            "code": escape(self.code),
        }


class Traceback(object):
    """Wraps a traceback."""

    def __init__(self, exc_type, exc_value, tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.tb = tb

        exception_type = exc_type.__name__
        if exc_type.__module__ not in {"builtins", "__builtin__", "exceptions"}:
            exception_type = exc_type.__module__ + "." + exception_type
        self.exception_type = exception_type

        self.groups = []
        memo = set()
        while True:
            self.groups.append(Group(exc_type, exc_value, tb))
            memo.add(id(exc_value))
            if PY2:
                break
            exc_value = exc_value.__cause__ or exc_value.__context__
            if exc_value is None or id(exc_value) in memo:
                break
            exc_type = type(exc_value)
            tb = exc_value.__traceback__
        self.groups.reverse()
        self.frames = [frame for group in self.groups for frame in group.frames]

    def filter_hidden_frames(self):
        """Remove the frames according to the paste spec."""
        for group in self.groups:
            group.filter_hidden_frames()

        self.frames[:] = [frame for group in self.groups for frame in group.frames]

    @property
    def is_syntax_error(self):
        """Is it a syntax error?"""
        return isinstance(self.exc_value, SyntaxError)

    @property
    def exception(self):
        """String representation of the final exception."""
        return self.groups[-1].exception

    def log(self, logfile=None):
        """Log the ASCII traceback into a file object."""
        if logfile is None:
            logfile = sys.stderr
        tb = self.plaintext.rstrip() + u"\n"
        logfile.write(to_native(tb, "utf-8", "replace"))

    def paste(self):
        """Create a paste and return the paste id."""
        data = json.dumps(
            {
                "description": "Werkzeug Internal Server Error",
                "public": False,
                "files": {"traceback.txt": {"content": self.plaintext}},
            }
        ).encode("utf-8")
        try:
            from urllib2 import urlopen
        except ImportError:
            from urllib.request import urlopen
        rv = urlopen("https://api.github.com/gists", data=data)
        resp = json.loads(rv.read().decode("utf-8"))
        rv.close()
        return {"url": resp["html_url"], "id": resp["id"]}

    def render_summary(self, include_title=True):
        """Render the traceback for the interactive console."""
        title = ""
        classes = ["traceback"]
        if not self.frames:
            classes.append("noframe-traceback")
            frames = []
        else:
            library_frames = sum(frame.is_library for frame in self.frames)
            mark_lib = 0 < library_frames < len(self.frames)
            frames = [group.render(mark_lib=mark_lib) for group in self.groups]

        if include_title:
            if self.is_syntax_error:
                title = u"Syntax Error"
            else:
                title = u"Traceback <em>(most recent call last)</em>:"

        if self.is_syntax_error:
            description_wrapper = u"<pre class=syntaxerror>%s</pre>"
        else:
            description_wrapper = u"<blockquote>%s</blockquote>"

        return SUMMARY_HTML % {
            "classes": u" ".join(classes),
            "title": u"<h3>%s</h3>" % title if title else u"",
            "frames": u"\n".join(frames),
            "description": description_wrapper % escape(self.exception),
        }

    def render_full(self, evalex=False, secret=None, evalex_trusted=True):
        """Render the Full HTML page with the traceback info."""
        exc = escape(self.exception)
        return PAGE_HTML % {
            "evalex": "true" if evalex else "false",
            "evalex_trusted": "true" if evalex_trusted else "false",
            "console": "false",
            "title": exc,
            "exception": exc,
            "exception_type": escape(self.exception_type),
            "summary": self.render_summary(include_title=False),
            "plaintext": escape(self.plaintext),
            "plaintext_cs": re.sub("-{2,}", "-", self.plaintext),
            "traceback_id": self.id,
            "secret": secret,
        }

    @cached_property
    def plaintext(self):
        return u"\n".join([group.render_text() for group in self.groups])

    @property
    def id(self):
        return id(self)


class Group(object):
    """A group of frames for an exception in a traceback. On Python 3,
    if the exception has a ``__cause__`` or ``__context__``, there are
    multiple exception groups.
    """

    def __init__(self, exc_type, exc_value, tb):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.info = None
        if not PY2:
            if exc_value.__cause__ is not None:
                self.info = (
                    u"The above exception was the direct cause of the"
                    u" following exception"
                )
            elif exc_value.__context__ is not None:
                self.info = (
                    u"During handling of the above exception, another"
                    u" exception occurred"
                )

        self.frames = []
        while tb is not None:
            self.frames.append(Frame(exc_type, exc_value, tb))
            tb = tb.tb_next

    def filter_hidden_frames(self):
        new_frames = []
        hidden = False

        for frame in self.frames:
            hide = frame.hide
            if hide in ("before", "before_and_this"):
                new_frames = []
                hidden = False
                if hide == "before_and_this":
                    continue
            elif hide in ("reset", "reset_and_this"):
                hidden = False
                if hide == "reset_and_this":
                    continue
            elif hide in ("after", "after_and_this"):
                hidden = True
                if hide == "after_and_this":
                    continue
            elif hide or hidden:
                continue
            new_frames.append(frame)

        # if we only have one frame and that frame is from the codeop
        # module, remove it.
        if len(new_frames) == 1 and self.frames[0].module == "codeop":
            del self.frames[:]

        # if the last frame is missing something went terrible wrong :(
        elif self.frames[-1] in new_frames:
            self.frames[:] = new_frames

    @property
    def exception(self):
        """String representation of the exception."""
        buf = traceback.format_exception_only(self.exc_type, self.exc_value)
        rv = "".join(buf).strip()
        return to_unicode(rv, "utf-8", "replace")

    def render(self, mark_lib=True):
        out = []
        if self.info is not None:
            out.append(u'<li><div class="exc-divider">%s:</div>' % self.info)
        for frame in self.frames:
            out.append(
                u"<li%s>%s"
                % (
                    u' title="%s"' % escape(frame.info) if frame.info else u"",
                    frame.render(mark_lib=mark_lib),
                )
            )
        return u"\n".join(out)

    def render_text(self):
        out = []
        if self.info is not None:
            out.append(u"\n%s:\n" % self.info)
        out.append(u"Traceback (most recent call last):")
        for frame in self.frames:
            out.append(frame.render_text())
        out.append(self.exception)
        return u"\n".join(out)


class Frame(object):
    """A single frame in a traceback."""

    def __init__(self, exc_type, exc_value, tb):
        self.lineno = tb.tb_lineno
        self.function_name = tb.tb_frame.f_code.co_name
        self.locals = tb.tb_frame.f_locals
        self.globals = tb.tb_frame.f_globals

        fn = inspect.getsourcefile(tb) or inspect.getfile(tb)
        if fn[-4:] in (".pyo", ".pyc"):
            fn = fn[:-1]
        # if it's a file on the file system resolve the real filename.
        if os.path.isfile(fn):
            fn = os.path.realpath(fn)
        self.filename = to_unicode(fn, get_filesystem_encoding())
        self.module = self.globals.get("__name__", self.locals.get("__name__"))
        self.loader = self.globals.get("__loader__", self.locals.get("__loader__"))
        self.code = tb.tb_frame.f_code

        # support for paste's traceback extensions
        self.hide = self.locals.get("__traceback_hide__", False)
        info = self.locals.get("__traceback_info__")
        if info is not None:
            info = to_unicode(info, "utf-8", "replace")
        self.info = info

    def render(self, mark_lib=True):
        """Render a single frame in a traceback."""
        return FRAME_HTML % {
            "id": self.id,
            "filename": escape(self.filename),
            "lineno": self.lineno,
            "function_name": escape(self.function_name),
            "lines": self.render_line_context(),
            "library": "library" if mark_lib and self.is_library else "",
        }

    @cached_property
    def is_library(self):
        return any(
            self.filename.startswith(path) for path in sysconfig.get_paths().values()
        )

    def render_text(self):
        return u'  File "%s", line %s, in %s\n    %s' % (
            self.filename,
            self.lineno,
            self.function_name,
            self.current_line.strip(),
        )

    def render_line_context(self):
        before, current, after = self.get_context_lines()
        rv = []

        def render_line(line, cls):
            line = line.expandtabs().rstrip()
            stripped_line = line.strip()
            prefix = len(line) - len(stripped_line)
            rv.append(
                '<pre class="line %s"><span class="ws">%s</span>%s</pre>'
                % (cls, " " * prefix, escape(stripped_line) or " ")
            )

        for line in before:
            render_line(line, "before")
        render_line(current, "current")
        for line in after:
            render_line(line, "after")

        return "\n".join(rv)

    def get_annotated_lines(self):
        """Helper function that returns lines with extra information."""
        lines = [Line(idx + 1, x) for idx, x in enumerate(self.sourcelines)]

        # find function definition and mark lines
        if hasattr(self.code, "co_firstlineno"):
            lineno = self.code.co_firstlineno - 1
            while lineno > 0:
                if _funcdef_re.match(lines[lineno].code):
                    break
                lineno -= 1
            try:
                offset = len(inspect.getblock([x.code + "\n" for x in lines[lineno:]]))
            except TokenError:
                offset = 0
            for line in lines[lineno : lineno + offset]:
                line.in_frame = True

        # mark current line
        try:
            lines[self.lineno - 1].current = True
        except IndexError:
            pass

        return lines

    def eval(self, code, mode="single"):
        """Evaluate code in the context of the frame."""
        if isinstance(code, string_types):
            if PY2 and isinstance(code, text_type):  # noqa
                code = UTF8_COOKIE + code.encode("utf-8")
            code = compile(code, "<interactive>", mode)
        return eval(code, self.globals, self.locals)

    @cached_property
    def sourcelines(self):
        """The sourcecode of the file as list of unicode strings."""
        # get sourcecode from loader or file
        source = None
        if self.loader is not None:
            try:
                if hasattr(self.loader, "get_source"):
                    source = self.loader.get_source(self.module)
                elif hasattr(self.loader, "get_source_by_code"):
                    source = self.loader.get_source_by_code(self.code)
            except Exception:
                # we munch the exception so that we don't cause troubles
                # if the loader is broken.
                pass

        if source is None:
            try:
                with open(
                    to_native(self.filename, get_filesystem_encoding()), mode="rb"
                ) as f:
                    source = f.read()
            except IOError:
                return []

        # already unicode?  return right away
        if isinstance(source, text_type):
            return source.splitlines()

        # yes. it should be ascii, but we don't want to reject too many
        # characters in the debugger if something breaks
        charset = "utf-8"
        if source.startswith(UTF8_COOKIE):
            source = source[3:]
        else:
            for idx, match in enumerate(_line_re.finditer(source)):
                match = _coding_re.search(match.group())
                if match is not None:
                    charset = match.group(1)
                    break
                if idx > 1:
                    break

        # on broken cookies we fall back to utf-8 too
        charset = to_native(charset)
        try:
            codecs.lookup(charset)
        except LookupError:
            charset = "utf-8"

        return source.decode(charset, "replace").splitlines()

    def get_context_lines(self, context=5):
        before = self.sourcelines[self.lineno - context - 1 : self.lineno - 1]
        past = self.sourcelines[self.lineno : self.lineno + context]
        return (before, self.current_line, past)

    @property
    def current_line(self):
        try:
            return self.sourcelines[self.lineno - 1]
        except IndexError:
            return u""

    @cached_property
    def console(self):
        return Console(self.globals, self.locals)

    @property
    def id(self):
        return id(self)



import base64

from logging import Logger
from xml.etree import cElementTree as eTree

from interactor.route.order.save.dto import (
    OrderSaveRequest,
    SaveOrderResponse,
)

from model.error_type import ExternalServicesErrorsType
from model.exception import ExternalServiceErrors
from report import CashPaymentReport
from services import PrinterService
from model.repository import PosRepositoryModel

from model.config import (
    Config,
    MapValidatePosId,
)
from validation_util import (
    check_if_business_day_is_opened,
    check_if_operator_is_logged,
    check_if_order_is_totaled,
    validate_pos_id,
    check_has_due_amount,
)


class OrderSaveInteractor(object):
    def __init__(
        self,
        logger,             # type: Logger
        config,             # type: Config
        pos_repository,     # type: PosRepositoryModel
        printer_service,    # type: PrinterService
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.printer_service = printer_service

    def execute(
        self,
        request,     # type: OrderSaveRequest
    ):
        # type: (...) -> SaveOrderResponse

        model_validate = MapValidatePosId(
            pos_id=request.pos_id,
            system_api_key=request.system_api_key,
            secure_api_key_list=self.config.api_key_mapping,
        )

        pos_id = validate_pos_id(
            request=model_validate,
        )

        model = self.pos_repository.get_model(
            pos_id=pos_id,
        )

        check_if_business_day_is_opened(
            repository=self.pos_repository,
            model=model,
        )
        check_if_operator_is_logged(
            repository=self.pos_repository,
            model=model,
        )
        check_if_order_is_totaled(
            repository=self.pos_repository,
            model=model,
        )
        check_has_due_amount(
            model=model,
            pos_id=pos_id,
            repository=self.pos_repository,
        )

        try:
            self.pos_repository.do_stored_order(
                model=model,
                pos_id=pos_id,
            )
        except Exception as _:
            self.logger.exception("Error stored order")
            raise ExternalServiceErrors(ExternalServicesErrorsType.StoredOrderError)

        return self._print_receipt(
            pos_id=pos_id,
            model=model,
            request=request,
        )

    def _print_receipt(
        self,
        pos_id,     # type: int
        model,      # type: eTree
        request,    # type: OrderSaveRequest
    ):
        # type: (...) -> SaveOrderResponse

        cash_report = CashPaymentReport(
            pos_id=pos_id,
        )

        current_order = self.pos_repository.get_current_order(
            model=model,
        )

        report_data = cash_report.generate(
            model=model,
            current_order=current_order,
        )
        if request.print_coupon:
            self.printer_service.send_print_data(
                model=model,
                response=report_data,
            )
            report_data = None

        report_data = base64.b64encode(report_data) if report_data is not None else ""

        return SaveOrderResponse(
            coupon=report_data,
        )



import json
import logging
from datetime import datetime, timedelta
from threading import Lock
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from helper import (
    remove_accents,
    format_cpf_cnpj,
    format_phone,
)
from mw_helper import (
    show_check_list_box_dialog,
    show_filtered_list_box_dialog,
    show_loyalty_dialog,
    show_message_dialog,
    show_message_options_dialog,


)
from posot import OrderTaker
from sysactions import (
    action,
    get_current_order,
    get_model,
    get_posot,
    show_keyboard,
    show_messagebox,
    translate_message,
    StopAction,
)

from actions.coupon.display_coupon import (
    get_benefits_from_order,
    remove_discounts,
)
from actions.customer import (
    get_customer_doc,
    get_customer_phone_number,
)
from actions.discount.apply_discount import apply_discount
from actions.loyalty.model import LoyaltyButtonEnum
from application.interactor import LoyaltyInteractor
from utilconfigurations import get_ask_document_pinpad
from utilfunctions import parse_loyalty_customer_info

loyalty_available_cache = None
last_cache_time = None  # type: Optional[int]
lock = Lock()
logger = logging.getLogger("PosActions")
title = "$INFORMATION"
loyalty_interactor = LoyaltyInteractor()


@action
def is_loyalty_component_available(
    pos_id,                         # type: str
    seconds_to_renew_cache='3600',  # type: str
):
    # type: (...) -> str

    global last_cache_time
    global loyalty_available_cache

    with lock:
        now = datetime.now()
        seconds_to_renew = int(seconds_to_renew_cache or '1800')
        renew_cache = not last_cache_time or now > last_cache_time + timedelta(seconds=seconds_to_renew)
        if loyalty_available_cache is None or renew_cache:
            loyalty_available_cache = loyalty_interactor.test_connection()
            last_cache_time = now
        return str(loyalty_available_cache).lower()


@action
def open_loyalty(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> bool

    while True:
        loyalty_properties = _get_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
            )

        has_customer_info = False
        if loyalty_properties:
            has_customer_info = _has_customer_info(
                loyalty_properties=loyalty_properties,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        if has_customer_info:
            welcome_message = _get_welcome_message(
                loyalty_properties=loyalty_properties,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
            loyalty_id_button = "$REMOVE_LOYALTY"
            is_disabled = ""
        else:
            welcome_message, loyalty_id_button = loyalty_interactor.get_welcome_message_button_loyalty()
            is_disabled = '#DISABLED'

        buttons = "{}|$AVAILABLE_COUPONS{}|$APPLIED_COUPONS{}|$CLOSE".format(
            loyalty_id_button,
            is_disabled,
            is_disabled,
        )
        resp = show_loyalty_dialog(
            pos_id=pos_id,
            buttons=buttons,
            message=welcome_message,
        )

        if resp in (None, LoyaltyButtonEnum.CANCEL):
            break

        _execute_loyalty_option(
            pos_id=pos_id,
            response=resp,
            customer_info=has_customer_info,
            order_id=order_id,
            table_id=table_id,
            table_seat=table_seat,
        )

    return True


def _execute_loyalty_option(
    pos_id,         # type: str
    response,       # type: int
    customer_info,  # type: bool
    order_id,       # type: str
    table_id,       # type: str
    table_seat,     # type: str

):
    # type: (...) -> None

    if response == LoyaltyButtonEnum.UPDATE_CPF:
        if customer_info:
            _remove_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
        else:
            set_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        return

    if response == LoyaltyButtonEnum.AVAILABLE_COUPONS:
        if customer_info:
            _show_available_coupons(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

        return

    if response == LoyaltyButtonEnum.APPLIED_COUPONS and customer_info:
        _show_applied_coupons(
            pos_id=pos_id,
            table_id=table_id,
            order_id=order_id,
            table_seat=table_seat,
        )


def _get_table_customer_doc(
        pos_id,         # type: str
        table_id,       # type: str
        table_seat,     # type: str
):
    # type: (...) -> Optional[str]

    from msgbusboundary import MsgBusTableService as tableService

    customer_doc = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_CUSTOMER_INFO",
        table_id=table_id,
    )
    if customer_doc:
        customer_doc = json.loads(s=customer_doc['LOYALTY_TABLE_CUSTOMER_INFO'])
        if customer_doc.get(table_seat, {}):
            customer_doc = customer_doc[table_seat].get('customer_doc', None)
        else:
            customer_doc = None

    return customer_doc


def set_loyalty_properties(
    pos_id,          # type: str
    table_id,        # type: str
    order_id,        # type: str
    table_seat,      # type: str
    loyalty_id="",     # type: Optional[str]
):
    # type: (...) -> None

    loyalty_type_id = loyalty_interactor.get_loyalty_id_type()
    if order_id not in [None, ""]:
        loyalty_properties, loyalty_id = build_loyalty_properties(
            pos_id=pos_id,
            order_id=order_id,
            table_seat=table_seat,
            loyalty_id=loyalty_id,
            loyalty_type_id=loyalty_type_id,
        )

        if not loyalty_id:
            return

        _persist_loyalty_properties_qsr(
            loyalty_properties=loyalty_properties,
            loyalty_id=loyalty_id,
            pos_id=pos_id,
            order_id=order_id,
        )

        if not loyalty_properties:
            coupons_to_apply_on_total = []
        else:
            _, coupons_to_apply_on_total, _ = _get_loyalty_coupons(
                loyalty_properties=loyalty_properties,
                order_id=order_id,
                validate_from_source=True,
            )

        is_order_totaled = _is_order_totaled(pos_id=pos_id)
        if is_order_totaled and coupons_to_apply_on_total:
            show_message_dialog(
                pos_id=pos_id,
                title=title,
                message="$AUTOMATIC_BENEFITS_TO_MANUAL",
            )

    elif table_id is not None:
        loyalty_properties, loyalty_id = build_loyalty_properties(
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
            loyalty_id=loyalty_id,
            loyalty_type_id=loyalty_type_id,
        )

        if not loyalty_id:
            return

        persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
        )


def _persist_loyalty_properties_qsr(
    loyalty_properties,             # type: Dict
    loyalty_id,                     # type: str
    pos_id,                         # type: str
    order_id,                       # type: str
    only_discounts_on_total=False,  # type: bool
):
    # type: (...) -> None

    props = dict(LOYALTY_ID=loyalty_id or "clear")
    customer_info = loyalty_properties.get("LOYALTY_CUSTOMER_INFO", {})
    if not only_discounts_on_total:
        props["LOYALTY_CUSTOMER_INFO"] = json.dumps(customer_info) if customer_info else "clear"

        benefits = loyalty_properties.get("LOYALTY_CUSTOMER_BENEFITS")
        props["LOYALTY_CUSTOMER_BENEFITS"] = json.dumps(benefits) if benefits else "clear"

    coupons_on_total = loyalty_properties.get("LOYALTY_COUPONS_TO_APPLY_ON_TOTAL", {})
    props["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"] = json.dumps(coupons_on_total) if coupons_on_total else "clear"

    if isinstance(customer_info, basestring):
        customer_info = json.loads(customer_info)
    customer = customer_info.get("customerInfo", {}).get("customer", {})
    name = customer.get("name", "").encode("utf-8")
    props["CUSTOMER_NAME"] = name or "clear"

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)
    pos_ot.setOrderCustomProperties(
        properties=props,
        orderid=order_id,
    )


def persist_loyalty_properties_ts(
    loyalty_properties,             # type: Dict
    pos_id,                         # type: str
    table_id,                       # type: str
    table_seat,                     # type: str
    only_discounts_on_total=False,  # type: bool
):
    # type: (...) -> None

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event

    table_custom_properties = {}
    if not only_discounts_on_total:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
        customer_info_value = json.dumps(customer_info) if customer_info else "{}"
        table_custom_properties['LOYALTY_TABLE_CUSTOMER_INFO'] = customer_info_value

        benefits = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_BENEFITS")
        benefits_value = json.dumps(benefits) if benefits else "{}"
        table_custom_properties['LOYALTY_TABLE_CUSTOMER_BENEFITS'] = benefits_value

        loyalty_ids = loyalty_properties.get("LOYALTY_TABLE_ID")
        loyalty_ids_value = json.dumps(loyalty_ids) if loyalty_ids else "{}"
        table_custom_properties['LOYALTY_TABLE_ID'] = loyalty_ids_value

    coupons_on_total = loyalty_properties.get("LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL")
    coupons_on_total_value = json.dumps(coupons_on_total) if coupons_on_total else "{}"
    table_custom_properties['LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL'] = coupons_on_total_value

    table_customer = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
    if isinstance(table_customer, basestring):
        table_customer = json.loads(table_customer)
    customer_info = table_customer.get(table_seat, {}).get("customerInfo", {})

    if not only_discounts_on_total and (table_seat == "table" or _is_tab(table_id=table_id)):
        customer = customer_info.get("customer", {})
        name = customer.get("name", "").encode("utf-8")
        table_custom_properties['CUSTOMER_NAME'] = name

    tableService().set_table_custom_properties(
        pos_id=pos_id,
        properties=json.dumps(table_custom_properties),
        table_id=table_id,
    )

    table_picture = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )

    create_table_modified_event(
        pos_id=pos_id,
        table_picture=table_picture,
    )


def _get_loyalty_properties(
    pos_id,         # type: str
    table_id=None,  # type: Optional[str]
    order_id=None,  # type: Optional[str]
):
    # type: (...) -> Optional[Dict]

    if order_id:
        return _get_loyalty_properties_for_qsr(
            pos_id=pos_id,
            order_id=order_id,
        )

    if table_id:
        return _get_loyalty_properties_for_ts(
            pos_id=pos_id,
            table_id=table_id,
        )

    return None


def _get_loyalty_properties_for_qsr(
    pos_id,    # type: str
    order_id,  # type: str
):
    # type: (...) -> Optional[Dict]

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order_xml = pos_ot.orderPicture(
        posid=pos_id,
        orderid=order_id,
    )
    order_picture_properties = eTree.XML(order_xml).find(".//CustomOrderProperties")

    customer_info = order_picture_properties.find(".//OrderProperty[@key='LOYALTY_CUSTOMER_INFO']")
    benefits = order_picture_properties.find(".//OrderProperty[@key='LOYALTY_CUSTOMER_BENEFITS']")
    coupons_on_total = order_picture_properties.find(".//OrderProperty[@key='LOYALTY_COUPONS_TO_APPLY_ON_TOTAL']")
    if customer_info is not None:
        customer_info = customer_info.get('value')
    if benefits is not None:
        benefits = benefits.get('value')
    if coupons_on_total is not None:
        coupons_on_total = coupons_on_total.get('value')

    customer_info = json.loads(customer_info) if customer_info else []
    benefits = json.loads(benefits) if benefits else []
    coupons_on_total = json.loads(coupons_on_total) if coupons_on_total else []

    loyalty_properties = {
        "LOYALTY_CUSTOMER_INFO": customer_info,
        "LOYALTY_CUSTOMER_BENEFITS": benefits,
        "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL": coupons_on_total,
    }
    return loyalty_properties


def parse_table_loyalty_customer_info(
    loyalty_ids,         # type: str
    customer_info,       # type: str
    benefits,            # type: str
    coupons_on_total,    # type: str
):
    # type: (...) -> dict

    if loyalty_ids:
        loyalty_ids = json.loads(loyalty_ids)

    if customer_info:
        customer_info = json.loads(customer_info)
        _parse_object_children(dictionary=customer_info)

    if benefits:
        benefits = json.loads(benefits)
        _parse_object_children(dictionary=benefits)

    if coupons_on_total:
        coupons_on_total = json.loads(coupons_on_total)
        _parse_object_children(dictionary=coupons_on_total)

    loyalty_properties = {
        "LOYALTY_TABLE_ID": loyalty_ids,
        "LOYALTY_TABLE_CUSTOMER_INFO": customer_info,
        "LOYALTY_TABLE_CUSTOMER_BENEFITS": benefits,
        "LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL": coupons_on_total,
    }
    return loyalty_properties


def _get_loyalty_properties_for_ts(
    pos_id,     # type: str
    table_id,   # type: str
):
    # type: (...) -> Optional[Dict]

    from msgbusboundary import MsgBusTableService as tableService

    table_xml = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )
    table_properties = eTree.XML(table_xml).find('Properties')
    loyalty_ids = table_properties.get('LOYALTY_TABLE_ID') or '{}'
    customer_info = table_properties.get('LOYALTY_TABLE_CUSTOMER_INFO') or '{}'
    benefits = table_properties.get('LOYALTY_TABLE_CUSTOMER_BENEFITS') or '{}'
    coupons_on_total = table_properties.get('LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL') or '{}'

    return parse_table_loyalty_customer_info(
        loyalty_ids=loyalty_ids,
        customer_info=customer_info,
        benefits=benefits,
        coupons_on_total=coupons_on_total,
    )


def _parse_object_children(
    dictionary,     # type: Dict
):
    # type: (...) -> None

    for key, value in dictionary.items():
        if isinstance(value, basestring):
            dictionary.update({key: json.loads(value)})


def _remove_loyalty_properties(
    pos_id,        # type: str
    table_id,      # type: str
    order_id,      # type: str
    table_seat,    # type: str
):
    # type: (...) -> None

    model = get_model(posid=pos_id)
    if _has_discounts_on_total(
        pos_id=pos_id,
        model=model,
    ):
        return

    if order_id not in [None, ""]:
        _persist_loyalty_properties_qsr(
            loyalty_properties={},
            loyalty_id="",
            order_id=order_id,
            pos_id=pos_id,
        )

    if table_id not in [None, ""]:
        loyalty_properties = _get_loyalty_properties(
            pos_id=pos_id,
            table_id=table_id,
            order_id=order_id,
        )

        for property_name, value in loyalty_properties.items():
            if table_seat in value:
                loyalty_properties[property_name].pop(table_seat)

        persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat
        )


def migrate_order_loyalty_info_to_table(
    pos_id,                 # type: str
    order,                  # type: eTree.XML
    table_id,               # type: str
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> None

    table_loyalty_properties = _get_loyalty_properties_for_ts(
        pos_id=pos_id,
        table_id=table_id,
    )

    loyalty_id = order.find(".//CustomOrderProperties/OrderProperty[@key='LOYALTY_ID']")
    if loyalty_id is not None:
        table_loyalty_properties["LOYALTY_TABLE_ID"].update({table_seat: loyalty_id.get("value")})

    customer_info = order.find(".//CustomOrderProperties/OrderProperty[@key='LOYALTY_CUSTOMER_INFO']")
    if customer_info is not None:
        table_loyalty_properties["LOYALTY_TABLE_CUSTOMER_INFO"].update({
            table_seat: json.loads(customer_info.get("value"))
        })

    customer_benefits = order.find(".//CustomOrderProperties/OrderProperty[@key='LOYALTY_CUSTOMER_BENEFITS']")

    if customer_benefits is not None:
        benefits = json.loads(customer_benefits.get("value"))
        table_loyalty_properties["LOYALTY_TABLE_CUSTOMER_BENEFITS"].update({table_seat: benefits})
        apply_on_total = [b.get("id") for b in benefits if b.get("automatic") == "ON_TOTAL"]

        applied_discounts = order.find(".//CustomOrderProperties/OrderProperty[@key='BENEFIT_LIST']")
        applied_discounts = json.loads(applied_discounts.get("value") or "{}") if applied_discounts is not None else {}
        applied_discounts = applied_discounts.get("benefitList", None)
        benefits = json.loads(applied_discounts or "[]") if applied_discounts else []

        applied_loyalty_benefits = []
        for benefit_detail in benefits:
            benefit = benefit_detail.get("benefit", {})
            benefit_id = benefit.get("id", None)
            if benefit.get("is_loyalty_benefit", False) and benefit_id and benefit_id not in apply_on_total:
                applied_loyalty_benefits.append(benefit_id)
        apply_on_total.extend(applied_loyalty_benefits)

        table_loyalty_properties["LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"].update({table_seat: apply_on_total})

    customer_name = order.find(".//CustomOrderProperties/OrderProperty[@key='CUSTOMER_NAME']")
    customer_name = customer_name.get("value") if customer_name is not None else None
    if table_seat == "table" and customer_name:
        table_loyalty_properties["CUSTOMER_NAME"] = customer_name

    from msgbusboundary import MsgBusTableService as tableService
    tableService().set_table_custom_properties(
        pos_id=pos_id,
        properties=json.dumps(table_loyalty_properties),
        table_id=table_id,
    )


def build_loyalty_properties(
    pos_id,                 # type: str
    loyalty_type_id,        # type: str
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat='table',     # type: Optional[str]
    loyalty_id="",          # type: Optional[str]
):
    # type: (...) -> Tuple[Dict, Optional[str]]

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)
    if loyalty_id in ["", None]:
        loyalty_id = get_user_input_by_type(
            pos_id=pos_id,
            pos_ot=pos_ot,
            loyalty_type_id=loyalty_type_id,
            default_doc="",
            stop_action=False,
        )
    if not loyalty_id:
        return loyalty_properties, None

    try:
        loyalty_customer_info = loyalty_interactor.get_loyalty_info(loyalty_id=loyalty_id)
    except LoyaltyInteractor.LoyaltyPartnerError:
        logger.exception("Partner error")
        show_messagebox(
            posid=pos_id,
            message="$LOYALTY_PARTNER_ERROR",
        )
        return loyalty_properties, None

    if loyalty_customer_info is None:
        show_messagebox(
            posid=pos_id,
            message="$LOYALTY_INFORMATION_NOT_FOUND",
        )
        return loyalty_properties, None

    result = parse_loyalty_customer_info(loyalty_customer_info=loyalty_customer_info)

    if not result:
        show_messagebox(
            posid=pos_id,
            message="$LOYALTY_INFORMATION_NOT_FOUND",
        )
        return loyalty_properties, None

    if order_id:
        loyalty_properties = result
        if _is_order_totaled(pos_id=pos_id):
            loyalty_properties["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"] = []

    elif table_id:
        result["LOYALTY_ID"] = str(loyalty_id)
        result = {
            key.replace("LOYALTY_", "LOYALTY_TABLE_"): value
            for key, value in result.items()
        }
        for key, value in loyalty_properties.items():
            if key not in result:
                continue
            value.update({table_seat: result[key]})

    return loyalty_properties, loyalty_id


def _has_customer_info(
    loyalty_properties,     # type: Dict
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> bool

    if order_id:
        customer_info = loyalty_properties.get("LOYALTY_CUSTOMER_INFO", {})
        return bool(customer_info)

    if table_id:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
        customer_info = customer_info.get(table_seat)
        return bool(customer_info)

    return False


def get_loyalty_identifier(
    customer,    # type: dict
):
    # type: (...) -> str

    loyalty_id_type = loyalty_interactor.get_loyalty_id_type()
    loyalty_id_type = "loyalty_id" if loyalty_id_type == "OTHER" else loyalty_id_type.lower()

    return customer.get(loyalty_id_type, "")


def _get_welcome_message(
    loyalty_properties,     # type: Dict
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> unicode

    if order_id:
        customer_info = loyalty_properties.get("LOYALTY_CUSTOMER_INFO", {})

    elif table_id:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
        customer_info = customer_info.get(table_seat, {})

    else:
        return "$LOYALTY_WELCOME_MESSAGE_CUSTOMER"

    customer_info = customer_info.get("customerInfo", {})
    customer = customer_info.get("customer", {})
    customer_name = customer.get("name", "-")
    customer_doc = format_cpf_cnpj(customer["document"]) if "document" in customer else "-"
    customer_phone = format_phone(customer["phone"]) if "phone" in customer else "-"

    additional_info = customer_info.get("additionalInfo", {})
    welcome_message = additional_info.get("welcomeMessage", "")

    message = "{}|{}|{}|{}|{}".format(
        "$WELCOME_CLIENT_LOYALTY_INFO",
        customer_name.encode("utf-8"),
        customer_doc,
        customer_phone,
        welcome_message.encode('utf-8'),
    )
    return message


def _show_available_coupons(
    pos_id,              # type: str
    table_id=None,       # type: Optional[str]
    order_id=None,       # type: Optional[str]
    table_seat="table",  # type: Optional[str]
):
    # type: (...) -> None

    if not (order_id or table_id):
        return

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    if not loyalty_properties:
        return

    is_order_totaled = _is_order_totaled(pos_id=pos_id)
    customer_benefits, coupons_to_apply_on_total, _ = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_id=table_id,
        table_seat=table_seat,
    )

    customer_benefits = customer_benefits or []
    coupons_to_apply_on_total = coupons_to_apply_on_total or []

    available_coupons = _get_available_coupons(
        pos_id=pos_id,
        customer_benefits=customer_benefits,
        coupons_to_apply_on_total=coupons_to_apply_on_total,
        is_order_totaled=is_order_totaled,
    )

    if not available_coupons:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$WITHOUT_ANY_LOYALTY_BENEFITS_TO_SHOW",
        )
        return

    options = [
        u"{} - {}".format(
            index + 1,
            coupon["coupon"]["name"]
        ).encode("utf-8")
        for index, coupon in enumerate(available_coupons)
    ]

    selected_index = show_filtered_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", mask="NOFILTER")

    if selected_index is None:
        return
    else:
        if apply_discount(
            pos_id=pos_id,
            selected_coupon=available_coupons[selected_index],
            is_loyalty_benefit=True,
        ):
            benefit_id = available_coupons[selected_index]["benefitId"]
            _apply_selected_benefit(
                loyalty_properties=loyalty_properties,
                pos_id=pos_id,
                benefit_id=benefit_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
            raise StopAction()


def _get_loyalty_coupons(
    loyalty_properties,             # type: Dict
    order_id=None,                  # type: Optional[str]
    table_id=None,                  # type: Optional[str]
    table_seat="table",             # type: Optional[str]
    current_used_key=False,         # type: Optional[bool]
    validate_from_source=False,     # type: Optional[bool]
):
    # type: (...) -> Tuple[Dict, List, str]

    customer_benefits = None
    coupons_to_apply_on_total = None

    if order_id:
        customer_benefits = loyalty_properties.get("LOYALTY_CUSTOMER_BENEFITS", [])
        coupons_to_apply_on_total = loyalty_properties.get("LOYALTY_COUPONS_TO_APPLY_ON_TOTAL", [])
        if not coupons_to_apply_on_total and validate_from_source:
            coupons_to_apply_on_total = [b.get("id") for b in customer_benefits if b.get("automatic") == "ON_TOTAL"]
        current_used_key = "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"

    elif table_id:
        customer_benefits = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_BENEFITS", {})
        customer_benefits = customer_benefits.get(table_seat)
        coupons_to_apply_on_total = loyalty_properties.get("LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL", {})
        coupons_to_apply_on_total = coupons_to_apply_on_total.get(table_seat)
        current_used_key = "LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"

    return customer_benefits, coupons_to_apply_on_total, current_used_key


def _show_applied_coupons(
    pos_id,                # type: str
    table_id=None,         # type: Optional[str]
    order_id=None,         # type: Optional[str]
    table_seat="table",    # type: Optional[str]
):
    # type: (...) -> None

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    customer_benefits, coupons_to_apply_on_total, current_used_key = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_id=table_id,
        table_seat=table_seat,
        current_used_key=True,
    )

    if not customer_benefits or not coupons_to_apply_on_total:
        show_messagebox(
            posid=pos_id,
            message="$ORDER_HAS_NO_APPLIED_DISCOUNT",
        )
        return

    coupon_options = _format_coupon_options(
        pos_id=pos_id,
        coupon_options=customer_benefits,
        coupons_to_apply_on_total=coupons_to_apply_on_total,
    )
    options = ["{0}".format(c["name"]) for c in coupon_options]

    if len(options) <= 1:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$WITHOUT_ANY_LOYALTY_BENEFITS_TO_SHOW",
        )
        return

    selected_indexes = show_check_list_box_dialog(
        pos_id=pos_id,
        options=options,
        title="$SELECT_ONE_COUPON",
        buttons="$REMOVE|$CANCEL",
        mask="NO_SALE_PANEL_IN_LOYALTY_BENEFITS",
    )

    if not len(selected_indexes):
        return

    is_order_totaled = _is_order_totaled(pos_id=pos_id)
    loyalty_property = loyalty_properties[current_used_key]
    if table_id and not is_order_totaled:
        loyalty_property = loyalty_property[table_seat]

    indexes = [int(i) for i in selected_indexes if int(i)]

    for coupon_item in coupon_options:
        if coupon_item.get("index", 0) not in indexes:
            continue

        coupon_id = coupon_item.get('id')

        removed_discount = _remove_loyalty_discount(
            pos_id=pos_id,
            coupon_item=coupon_item,
            is_order_totaled=is_order_totaled,
        )
        if len(loyalty_property) > 0 and (removed_discount is True or not is_order_totaled):
            loyalty_property.remove(coupon_id)

    _apply_selected_benefit(
        loyalty_properties=loyalty_properties,
        pos_id=pos_id,
        benefit_id=None,
        table_id=table_id,
        order_id=order_id,
        table_seat=table_seat,
    )

    if is_order_totaled:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$ADD_ITEM_DISCOUNT_NOT_REMOVED",
        )


def _format_coupon_options(
    pos_id,                     # type: str
    coupon_options,             # type: Dict
    coupons_to_apply_on_total,  # type: List
):
    # type: (...) -> List

    model = get_model(posid=pos_id)
    all_coupons = translate_message(model, "ALL")
    applied_coupons = []
    benefit_index = 1

    order_benefits_ids = _get_order_benefits_ids_dict(model=model)
    is_order_totaled = _is_order_totaled(pos_id=pos_id)

    for coupon in coupon_options:
        benefit_id = coupon.get('id')
        benefit_key = order_benefits_ids[benefit_id] if benefit_id in order_benefits_ids.keys() else None
        add_item = _has_item_to_add(coupon=coupon)
        is_coupon_applied = benefit_id not in order_benefits_ids.keys()

        if (benefit_id not in coupons_to_apply_on_total) or (is_order_totaled and is_coupon_applied):
            continue

        disable = False
        description = (coupon.get("userDescription") or coupon.get("name") or coupon.get("id") or "").encode("UTF-8")
        name = "{}##{}##{}".format(description, 'null', disable)
        data = {
            "index": benefit_index,
            "id": benefit_id,
            "name": name,
            "benefit_key": benefit_key,
            "add_item": add_item
        }
        applied_coupons.append(data)
        benefit_index += 1

    name = "{}##{}##False".format(all_coupons, 'null')
    data = {"index": 0, "id": None, "name": name}
    applied_coupons.append(data)
    applied_coupons.sort(key=lambda x: (x["index"]))
    return applied_coupons


def _apply_selected_benefit(
    loyalty_properties,  # type: Dict
    pos_id,              # type: str
    benefit_id,          # type: Optional[str]
    table_id=None,       # type: Optional[str]
    order_id=None,       # type: Optional[str]
    table_seat="table",  # type: Optional[str]
):
    # type: (...) -> None

    if order_id:
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"]
        if benefit_id:
            coupons_to_apply_on_total.append(benefit_id)

        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        pos_ot.setOrderCustomProperty(
            "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL",
            value=json.dumps(coupons_to_apply_on_total),
            orderid=order_id
        )

    elif table_id:
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"]
        if benefit_id:
            if table_seat in coupons_to_apply_on_total:
                coupons_to_apply_on_total[table_seat].append(benefit_id)
            else:
                coupons_to_apply_on_total[table_seat] = [benefit_id]

        persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
            only_discounts_on_total=True,
        )


def _is_tab(
    table_id,   # str
):
    # type: (...) -> bool

    return 'TAB' in table_id


def _get_available_coupons(
    pos_id,                     # type: str
    customer_benefits,          # type: List
    coupons_to_apply_on_total,  # type: List
    is_order_totaled,           # type: bool
):
    # type: (...) -> List[Dict]

    benefit_list = []
    model = get_model(posid=pos_id)
    order_benefits = _get_applied_benefit_on_order(model=model)
    order_benefits_ids = [benefit["benefit"]["id"] for benefit in order_benefits]
    automatic_benefit = remove_accents(translate_message(model, "AUTOMATIC_BENEFIT"))

    for index, benefit in enumerate(customer_benefits):
        applied_benefit = _validate_applied_benefit(
            benefit=benefit,
            coupons_to_apply_on_total=coupons_to_apply_on_total,
            order_benefits_ids=order_benefits_ids,
            is_order_totaled=is_order_totaled,
        )

        if applied_benefit:
            continue

        name = str(benefit.get("userDescription", benefit.get("name", benefit.get("id", ""))).encode("utf-8"))
        is_automatic = benefit.get('automatic', None) not in [None, 'DISABLED']

        if name.startswith("$"):
            params = name.split("|", 1)[1] if "|" in name else ""
            message = name.split("|", 1)[0] if params else name
            name = translate_message(model, message[1:], *(params.split("|")))

        if is_order_totaled and is_automatic:
            name = name + " " + ("({0})".format(automatic_benefit))

        benefit_item = {
            "coupon": benefit,
            "name": name,
            "benefitId": benefit.get("id"),
        }
        benefit_list.append(benefit_item)

    return benefit_list


def _is_order_totaled(
    pos_id=None,     # type: str
    order=None,      # type: Element
):
    # type: (...) -> bool

    if not order:
        model = get_model(posid=pos_id)
        order = get_current_order(model=model)
    return order.get('state') == "TOTALED"


def _get_order_benefits_ids_dict(
    model,     # type: Element
):
    # type: (...) -> Dict

    order_benefits = _get_applied_benefit_on_order(model=model)
    benefits_dict = {}
    for _, benefit in enumerate(order_benefits):
        benefits_dict.update({benefit["benefit"]["id"]: benefit["benefit_key"]})
    return benefits_dict


def _has_item_to_add(
    coupon,     # type: Dict
):
    # type: (...) -> Optional[bool]

    coupon_actions = coupon.get("actions")
    for coupon_action in coupon_actions.values():
        if coupon_action.get("itemsToAdd"):
            return True


def _remove_loyalty_discount(
    pos_id,            # type: str
    coupon_item,       # type: Dict
    is_order_totaled,  # type: bool
):
    # type: (...) -> Optional[bool]

    coupon_key = coupon_item.get('benefit_key')
    remove_discounts(
        pos_id=pos_id,
        coupon_ids=coupon_key,
    )
    return True


def _validate_applied_benefit(
    benefit,                    # type: Dict
    coupons_to_apply_on_total,  # type: List
    order_benefits_ids,         # type: List
    is_order_totaled,           # type: bool
):
    # type: (...) -> bool

    applied_benefit = benefit.get("id") in coupons_to_apply_on_total
    benefit_applied_in_order = benefit.get("id") in order_benefits_ids
    if is_order_totaled:
        applied_benefit = applied_benefit and benefit_applied_in_order

    return applied_benefit


def _get_applied_benefit_on_order(
    model,     # type: Element
):
    # type: (...) -> Dict

    pos_ot = get_posot(model=model)
    order = get_current_order(model=model)
    order_benefits = get_benefits_from_order(
        pos_ot=pos_ot,
        order=order,
    )
    return order_benefits


def _has_discounts_on_total(
    pos_id,     # type: str
    model,      # type: Element
):
    # type: (...) -> bool

    order_benefits = _get_applied_benefit_on_order(model=model)
    is_order_totaled = _is_order_totaled(pos_id=pos_id)
    if order_benefits and is_order_totaled:
        show_message_dialog(
            pos_id=pos_id,
            title=title,
            message="$CANNOT_REMOVE_LOYALTY_ON_TOTALED_SCREEN",
        )
        return True


def get_user_input_by_type(
    pos_id,             # type: str
    pos_ot,             # type: OrderTaker
    loyalty_type_id,    # type: str
    default_doc,        # type: str
    stop_action,        # type: bool
):
    # type: (...) -> str

    document_pin_pad = get_ask_document_pinpad(pos_id=pos_id)
    if loyalty_type_id == "DOCUMENT":
        return get_customer_doc(
            pos_id=pos_id,
            pos_ot=pos_ot,
            default_doc=default_doc,
            ask_document_pinpad=document_pin_pad,
            stop_action=stop_action,
        )

    if loyalty_type_id == "PHONE":
        return get_customer_phone_number(
            pos_id=pos_id,
            stop_action=stop_action,
        )

    if loyalty_type_id == "OTHER":
        loyalty_id = show_keyboard(
            posid=pos_id,
            message="$ENTER_LOYALTY_ID",
            defvalue="",
            title="",
        )
        if loyalty_id is None or loyalty_id.strip() == "":
            return ""

        return loyalty_id.translate(None, '-+[]/;.|<>:?{}=_()*&^%$#@!,').upper()


def get_table_loyalty_customer_attribute(
    pos_id,                 # type: str
    table_id,               # type: str
    attribute="document",   # type: Optional[str]
):
    # type: (...) -> str

    if table_id is None:
        return ''

    from msgbusboundary import MsgBusTableService as tableService

    loyalty_info = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_CUSTOMER_INFO",
        table_id=table_id,
    )
    if loyalty_info:
        loyalty_info = json.loads(loyalty_info.get("LOYALTY_TABLE_CUSTOMER_INFO"))
        customer = loyalty_info.get("table", {}).get("customerInfo", {}).get("customer", {})
        return customer.get(attribute, "")
    return ""


def get_order_loyalty_customer_attribute(
    pos_id,                 # type: str
    order_picture,          # type: Element
    attribute="document",   # type: Optional[str]
):
    # type: (...) -> str

    loyalty_info = order_picture.find(".//OrderProperty[@key='LOYALTY_CUSTOMER_INFO']")
    table_id = order_picture.find(".//OrderProperty[@key='TABLE_ID']")
    if loyalty_info is not None:
        return json.loads(loyalty_info.get("value")).get("customerInfo", {}).get("customer", {}).get(attribute, "")
    elif table_id is not None:
        return get_table_loyalty_customer_attribute(
            pos_id=pos_id,
            table_id=table_id.get("value"),
            attribute=attribute,
        )
    return ""


def get_document_from_loyalty_or_customer_response(
    pos_id,                     # type: str
    pos_ot,                     # type: OrderTaker
    ask_document_pinpad=False,  # type: bool
    document=None,              # type: Optional[str]
    stop_action=True,           # type: Optional[bool]
):
    # type: (...) -> Optional[str]

    if document is not None:
        return None

    order_picture = eTree.XML(pos_ot.orderPicture(posid=pos_id))
    property_document = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='CUSTOMER_DOC']")
    if property_document is not None and property_document.get("value"):
        return None

    is_valid_document = get_order_loyalty_customer_attribute(
        pos_id=pos_id,
        order_picture=order_picture,
        attribute="isValidDocument",
    )

    if is_valid_document:
        loyalty_document = get_order_loyalty_customer_attribute(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        buttons = "$CONFIRM|$OTHER|$WITHOUT_DOCUMENT"
        message = "$CONFIRM_CPF_CNPJ|{0}".format(format_cpf_cnpj(value=loyalty_document))
        resp = show_message_options_dialog(
            pos_id=pos_id,
            buttons=buttons,
            title="$CONFIRMATION",
            message=message,
        )
        if resp == 2 or resp is None:
            return ""
        if resp == 0:
            return loyalty_document

    return get_customer_doc(
        pos_id=pos_id,
        pos_ot=pos_ot,
        ask_document_pinpad=ask_document_pinpad,
        stop_action=stop_action,
    ) or ""



# encoding: utf-8
import json

from typing import (
    Any,
    Optional
)

from hourlysalesreport import HourlySalesReport
from hourlysalesreport.repository import OrderRepository as HourlySalesOrderRepository


class ReportGenerator(object):
    @staticmethod
    def generate(
        mb_context,                     # type: Any
        pos_id,                         # type: str
        pos_list,                       # type: str
        operator_id,                    # type: str
        period,                         # type: str
        by_created_order_time=False     # type: Optional[bool]
    ):
        # type: (...) -> str

        pos_list = json.loads(pos_list)
        hourly_sales_report_generator = HourlySalesReport(
            order_repository=HourlySalesOrderRepository(
                mb_context=mb_context,
                pos_list=pos_list
            ),
            by_created_order_time=by_created_order_time
        )

        return hourly_sales_report_generator.generate_hourly_sales_report(
            pos_id=pos_id,
            user_id=operator_id,
            period=period
        )



from abc import (
    ABCMeta,
    abstractmethod,
)

from posot import OrderTaker


class LoyaltyServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_customer_info(
        self,
        pos_ot,         # type: OrderTaker
        customer_id,    # type: str
    ):
        # type: (...) -> SearchCustomerInfoResponse

        raise NotImplementedError



# -*- coding: utf-8 -*-
from dialogmanager import DialogManager
from messagehandler import MessageHandlerBuilder
from messagepaymentprocessor import (
    CancelPaymentProcessor,
    CancelPaymentOrderProcessor,
    ConfirmFinishPaymentProcessor,
    FinishPaymentProcessor,
    GetDocumentProcessor,
    GetElectronicTypesProcessor,
    TestConnectionProcessor,
    StartPaymentProcessor,
    GetNameProcessor,
    CacheProcessor,
)
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from messageprocessorutil import LoggingProcessorCallback
from messagetoken import MessageToken
from paymentrouter import PaymentRouter
from typing import Optional
from util import Cache


class ElectronicPaymentMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, message_bus, config, logger):
        self.message_bus = message_bus
        self.config = config
        self.logger = logger

        self.message_handler = None
        self.dialog_manager = None  # type: Optional[DialogManager]
        self.payment_router = None  # type: Optional[PaymentRouter]
        self.cache = None           # type: Optional[Cache]

    def build_singletons(
        self,
    ):
        # type: (...) -> None

        self.dialog_manager = DialogManager(
            message_bus=self.message_bus,
            logger=self.logger,
        )

        self.cache = Cache(
            message_bus=self.message_bus,
            dialog_manager=self.dialog_manager,
            logger=self.logger,
        )

        self.payment_router = PaymentRouter(
            message_bus=self.message_bus,
            dialog_manager=self.dialog_manager,
            config=self.config,
            logger=self.logger,
            cache=self.cache,
        )

    def build_message_handler(self, _, __):
        event_processors = {}
        message_processors = {
            MessageToken.TK_ELECTRONIC_PAYMENT_ADD_PAYMENT: StartPaymentProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT: CancelPaymentProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT_ORDER: CancelPaymentOrderProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT: FinishPaymentProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_CONFIRM_FINISH: ConfirmFinishPaymentProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_CONNECTIVITY_TEST: TestConnectionProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_GET_ELECTRONIC_TYPES: GetElectronicTypesProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_GET_DOCUMENT: GetDocumentProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_GET_NAME: GetNameProcessor(self.payment_router),
            MessageToken.TK_ELECTRONIC_PAYMENT_SEND_RESPONSE: self.dialog_manager,
            MessageToken.TK_ELECTRONIC_PAYMENT_CACHE: CacheProcessor(
                payment_router=self.payment_router,
                cache=self.cache,
            ),
        }
        self.message_handler = MessageProcessorMessageHandler(
            event_processors,
            message_processors,
            DefaultMessageProcessorExecutorFactory([LoggingProcessorCallback(self.logger)]),
            None
        )
        return self.message_handler

    def destroy_message_handler(self, message_handler):
        pass  # no special handling necessary to destroy message handlers

    def destroy_singletons(self):
        self.payment_router.stop_processors()



from _CashReportBodyDto import CashReportBodyDto  # noqa


class LogoutSummaryBodyDto(object):
    def __init__(self, authorizer, login_time, report_body):
        # type: (float, unicode, unicode, CashReportBodyDto) -> ()
        self.authorizer = authorizer
        self.login_time = login_time
        self.report_body = report_body



# -*- coding: utf-8 -*-

"""

pyBarcode
=========

This package provides a simple way to create standard barcodes.
It needs no external packages to be installed, the barcodes are
created as SVG objects. If PIL (Python Imaging Library) is
installed, the barcodes can also be rendered as images (all
formats supported by PIL).
"""

__project__ = 'pyBarcode'
__author__ = 'Thorsten Weimann'
__copyright__ = '2010-2013, ' + __author__
__author_email__ = 'weimann@ymail.com'
__description__ = ('Create standard barcodes with Python. No external '
                   'modules needed (optional PIL support included).')
__version__ = '0.8'
__release__ = '{version}beta1'.format(version=__version__)
__license__ = 'MIT'
__url__ = 'https://bitbucket.org/whitie/python-barcode/'
__classifiers__ = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Multimedia :: Graphics',
]


from barcode.errors import BarcodeNotFoundError
from barcode.codex import Code39, PZN, Code128
from barcode.ean import EAN8, EAN13, JAN
from barcode.isxn import ISBN10, ISBN13, ISSN
from barcode.upc import UPCA
from barcode.itf import ITF

try:
    _strbase = basestring  # lint:ok
except NameError:
    _strbase = str


__BARCODE_MAP = dict(
    ean8=EAN8,
    ean13=EAN13,
    ean=EAN13,
    gtin=EAN13,
    jan=JAN,
    upc=UPCA,
    upca=UPCA,
    isbn=ISBN13,
    isbn13=ISBN13,
    gs1=ISBN13,
    isbn10=ISBN10,
    issn=ISSN,
    code39=Code39,
    pzn=PZN,
    code128=Code128,
    itf=ITF,
)

PROVIDED_BARCODES = list(__BARCODE_MAP.keys())
PROVIDED_BARCODES.sort()


def get(name, code=None, writer=None):
    try:
        barcode = __BARCODE_MAP[name.lower()]
    except KeyError:
        raise BarcodeNotFoundError('The barcode {0!r} you requested is not '
                                   'known.'.format(name))
    if code is not None:
        return barcode(code, writer)
    else:
        return barcode


def get_class(name):
    return get_barcode(name)


def generate(name, code, writer=None, output=None, writer_options=None):
    options = writer_options or {}
    barcode = get_barcode(name, code, writer)
    if isinstance(output, _strbase):
        fullname = barcode.save(output, options)
        return fullname
    else:
        barcode.write(output, options)


get_barcode = get
get_barcode_class = get_class



from application.interactor import (
    BenefitRetryInteractor,
    BurnBenefitInteractor,
    GetBenefitInfoInteractor,
    UnlockBenefitInteractor,
    OrderModifiedPaidInteractor,
    OrderModifiedVoidInteractor,
    OrderModifiedVoidLineInteractor,
    TestConnectionInteractor
)
from application.interactor.apply.apply_benefit import ApplyBenefitInteractor
from application.interactor.apply.apply_coupon import ApplyCouponInteractor
from application.interactor.apply.service import BenefitApplierService
from application.manager import BenefitRetryManager
from application.model import (
    ListenedTokens,
    ListenedEvents,
    BenefitAppliers,
)
from application.model.configuration import Configurations
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.model.repository.benefit_applier.exception import (
    CouponCustomError,
    CouponNotFound,
    CouponAlreadyUsed,
    CouponLocked,
    BenefitApplierNotFound,
)
from application.processor import (
    GetBenefitInfoProcessor,
    OrderModifiedPaidProcessor,
    OrderModifiedVoidProcessor,
    OrderModifiedVoidLineProcessor,
    NotMappedProcessor,
    TestConnectionProcessor
)
from application.processor.apply.apply_benefit._ApplyBenefitProcessor import ApplyBenefitProcessor
from application.processor.apply.apply_coupon import ApplyCouponProcessor
from application.repository.pos import DBRepository
from messagebus import Message, Event
from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    MessageProcessorMessageHandler,
    DefaultMessageProcessorExecutorFactory,
)
from messageprocessorutil import LoggingProcessorCallback
from typing import Union, Dict

from application.services import TableService


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(
        self,                               # type: ComponentMessageHandlerBuilder
        configs,                            # type: Configurations
        db_repository,                      # type: DBRepository
        benefit_appliers_repositories,      # type: Dict[BenefitAppliers: DefaultBenefitApplierRepository]
        table_service,                      # type: TableService
    ):
        # type: (...) -> None

        self.logger = configs.logger
        self.configs = configs
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories
        self.table_service = table_service

        self.benefit_retry_manager = None  # type: Union[None, BenefitRetryManager]
        self.message_handler = None  # type: Union[None, MessageProcessorMessageHandler]

    def build_singletons(self):
        # type: () -> None

        default_params = [
            self.configs,
            self.db_repository,
            self.benefit_appliers_repositories,
        ]

        benefit_applier_service = BenefitApplierService(
            benefit_appliers_repositories=self.benefit_appliers_repositories,
            benefit_applier_tender_ids=self.configs.benefit_applier_tender_ids,
            db_repository=self.db_repository,
        )
        burn_benefit_interactor = BurnBenefitInteractor(
            configs=self.configs,
            db_repository=self.db_repository,
            benefit_appliers_repositories=self.benefit_appliers_repositories,
            table_service=self.table_service,
        )
        get_benefit_info_interactor = GetBenefitInfoInteractor(*default_params)
        apply_coupon_interactor = ApplyCouponInteractor(
            benefit_applier_service=benefit_applier_service
        )
        apply_benefit_interactor = ApplyBenefitInteractor(
            benefit_applier_service=benefit_applier_service,
        )
        unlock_benefit_interactor = UnlockBenefitInteractor(
            configs=self.configs,
            db_repository=self.db_repository,
            benefit_appliers_repositories=self.benefit_appliers_repositories,
            table_service=self.table_service,
        )
        test_connection_interactor = TestConnectionInteractor(self.configs)

        order_modified_paid_interactor = OrderModifiedPaidInteractor(burn_benefit_interactor, *default_params)
        order_modified_void_interactor = OrderModifiedVoidInteractor(
            unlock_benefit_interactor=unlock_benefit_interactor,
            configs=self.configs,
            db_repository=self.db_repository,
            benefit_appliers_repositories=self.benefit_appliers_repositories,
            table_service=self.table_service,
        )
        order_modified_void_line_interactor = OrderModifiedVoidLineInteractor(
            unlock_benefit_interactor,
            *default_params
        )

        get_benefit_info_processor = GetBenefitInfoProcessor(get_benefit_info_interactor)
        apply_coupon_processor = ApplyCouponProcessor(interactor=apply_coupon_interactor)
        apply_benefit_processor = ApplyBenefitProcessor(interactor=apply_benefit_interactor)

        order_modified_paid_processor = OrderModifiedPaidProcessor(order_modified_paid_interactor)
        order_modified_void_processor = OrderModifiedVoidProcessor(order_modified_void_interactor)
        order_modified_void_line_processor = OrderModifiedVoidLineProcessor(order_modified_void_line_interactor)
        not_mapped_processor = NotMappedProcessor()

        test_connection_processor = TestConnectionProcessor(test_connection_interactor)

        messages_processors = {
            ListenedTokens.TK_NCOUPON_APPLY_COUPON.value: apply_coupon_processor,
            ListenedTokens.TK_NCOUPON_APPLY_BENEFIT.value: apply_benefit_processor,
            ListenedTokens.TK_NCOUPON_GET_COUPON_INFO.value: get_benefit_info_processor,
            ListenedTokens.TK_NCOUPON_TEST_CONNECTION.value: test_connection_processor
        }

        events_processors = {
            ListenedEvents.ORDER_MODIFIED_PAID.value: order_modified_paid_processor,
            ListenedEvents.ORDER_MODIFIED_VOID.value: order_modified_void_processor,
            ListenedEvents.ORDER_MODIFIED_VOID_LINE.value: order_modified_void_line_processor,
            ListenedEvents.ORDER_MODIFIED.value: not_mapped_processor
        }

        callbacks = [
            LoggingProcessorCallback(
                logger=self.logger,
                ignored_exceptions=[
                    CouponCustomError,
                    CouponNotFound,
                    CouponAlreadyUsed,
                    CouponLocked,
                    BenefitApplierNotFound,
                ]
            )
        ]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=events_processors,
            message_processors=messages_processors,
            executor_factory=DefaultMessageProcessorExecutorFactory(callbacks),
        )

        self.benefit_retry_manager = self._build_benefits_retry_manager(
            burn_benefit_interactor=burn_benefit_interactor,
            unlock_benefit_interactor=unlock_benefit_interactor,
        )
        self.benefit_retry_manager.start()

    def build_message_handler(self, message, event):
        # type: (Message, Event) -> None

        return self.message_handler

    def destroy_singletons(self):
        # type: () -> None

        self.benefit_retry_manager.stop()

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None

        return

    def _build_benefits_retry_manager(self, burn_benefit_interactor, unlock_benefit_interactor):
        # type: (BurnBenefitInteractor, UnlockBenefitInteractor) -> BenefitRetryManager

        interactor = BenefitRetryInteractor(
            self.configs,
            self.db_repository,
            burn_benefit_interactor,
            unlock_benefit_interactor
        )
        return BenefitRetryManager(self.configs, interactor)



from abc import ABCMeta, abstractmethod

from ._Waiter import Waiter


class WaiterRetriever(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_waiter(self, sync_id):
        # type: (int) -> Waiter
        raise NotImplementedError()



from json import JSONEncoder

from report.command import NewLineCommand

from ._Report import Report
from ._Part import Part
from .command import (
    AlignCommand,
    BoldCommand,
    FontCommand,
    RepeatCommand,
    CutPaperCommand
)


class ReportJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Report):
            return {
                "key": Report.key,
                "parts": o.get_parts()
            }

        if isinstance(o, Part):
            return {
                "text": o.text,
                "commands": o.commands
            }

        if isinstance(o, AlignCommand):
            return {
                "type": "align",
                "alignment": o.alignment
            }

        if isinstance(o, BoldCommand):
            return {
                "type": "bold",
                "bold": o.activate
            }

        if isinstance(o, FontCommand):
            return {
                "type": "font",
                "font": o.font
            }

        if isinstance(o, RepeatCommand):
            return {
                "type": "repeat",
                "text": o.text,
                "count": o.count
            }

        if isinstance(o, NewLineCommand):
            return {
                "type": "newLine"
            }

        if isinstance(o, CutPaperCommand):
            return {
                "type": "cutPaper"
            }

        return JSONEncoder.default(self, o)



# Copyright Jonathan Hartley 2013. BSD 3-Clause license, see LICENSE file.
import sys
from unittest import TestCase, main

from ..ansitowin32 import StreamWrapper, AnsiToWin32
from .utils import pycharm, replace_by, replace_original_by, StreamTTY, StreamNonTTY


def is_a_tty(stream):
    return StreamWrapper(stream, None).isatty()

class IsattyTest(TestCase):

    def test_TTY(self):
        tty = StreamTTY()
        self.assertTrue(is_a_tty(tty))
        with pycharm():
            self.assertTrue(is_a_tty(tty))

    def test_nonTTY(self):
        non_tty = StreamNonTTY()
        self.assertFalse(is_a_tty(non_tty))
        with pycharm():
            self.assertFalse(is_a_tty(non_tty))

    def test_withPycharm(self):
        with pycharm():
            self.assertTrue(is_a_tty(sys.stderr))
            self.assertTrue(is_a_tty(sys.stdout))

    def test_withPycharmTTYOverride(self):
        tty = StreamTTY()
        with pycharm(), replace_by(tty):
            self.assertTrue(is_a_tty(tty))

    def test_withPycharmNonTTYOverride(self):
        non_tty = StreamNonTTY()
        with pycharm(), replace_by(non_tty):
            self.assertFalse(is_a_tty(non_tty))

    def test_withPycharmNoneOverride(self):
        with pycharm():
            with replace_by(None), replace_original_by(None):
                self.assertFalse(is_a_tty(None))
                self.assertFalse(is_a_tty(StreamNonTTY()))
                self.assertTrue(is_a_tty(StreamTTY()))

    def test_withPycharmStreamWrapped(self):
        with pycharm():
            self.assertTrue(AnsiToWin32(StreamTTY()).stream.isatty())
            self.assertFalse(AnsiToWin32(StreamNonTTY()).stream.isatty())
            self.assertTrue(AnsiToWin32(sys.stdout).stream.isatty())
            self.assertTrue(AnsiToWin32(sys.stderr).stream.isatty())


if __name__ == '__main__':
    main()



from logging import Logger

from typing import List


class Configurations(object):
    def __init__(
        self,
        required_services=None,
        service_name=None,
        persistence_name=None,
        logger=None,
        supported_benefit_appliers=None,
        retry_frequency=None,
        max_retry_number=None,
        retry_min_time_to_try_again=None,
        benefit_applier_tender_ids=None
    ):
        # type: (str, str, str, Logger, List[str], int, int, int, dict) -> None

        self.required_services = required_services
        self.service_name = service_name
        self.persistence_name = persistence_name
        self.logger = logger
        self.supported_benefit_appliers = supported_benefit_appliers
        self.retry_frequency = retry_frequency
        self.max_retry_quantity = max_retry_number
        self.retry_min_time_to_try_again = retry_min_time_to_try_again
        self.benefit_applier_tender_ids = benefit_applier_tender_ids



from ._LoyaltyToken import LoyaltyToken
from messagebus import TokenCreator, TokenPriority


class LoyaltyRequestToken(LoyaltyToken):
    TK_LOYALTY_GET_AND_LOCK_VOUCHER = LoyaltyToken.create_token("2")
    TK_LOYALTY_BURN_VOUCHER = LoyaltyToken.create_token("3")
    TK_LOYALTY_UNLOCK_VOUCHER = LoyaltyToken.create_token("4")
    TK_LOYALTY_LOCK_BENEFIT = LoyaltyToken.create_token("6")



from typing import List

from devicemanager.interactor.dto import DeviceStatusDto
from devicemanager.service import GetDevicesService


class GetDevicesInteractor(object):
    def __init__(
        self,
        get_devices_service,  # type: GetDevicesService
    ):
        # type: (...) -> None

        self.get_devices_service = get_devices_service

    def get_devices(
        self,  # type: GetDevicesInteractor
    ):
        # type: (...) -> List[DeviceStatusDto]

        return self.get_devices_service.get_devices()



import persistence
from msgbus import MBEasyContext
from pos_model import PosConnection


class BaseRepository(object):
    def __init__(self, mbcontext, dbname = None):
        # type: (MBEasyContext, str) -> None
        self.mbcontext = mbcontext
        self.dbname = dbname
        self.conn = None  # type: persistence.Connection

    def __enter__(self):
        self.conn = persistence.Driver().open(self.mbcontext)
        if self.dbname is not None:
            self.conn.set_dbname(self.dbname)

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self.conn is not None:
            self.conn.close()


class BaseTransactionRepository(BaseRepository):
    def __init__(self, mbcontext, dbname = None):
        # type: (MBEasyContext, str) -> None
        super(BaseTransactionRepository, self).__init__(mbcontext, dbname)

    def __enter__(self):
        self.open_connection()
        self.begin_transaction()

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_value is None:
            self.commit()
        else:
            self.rollback()

        self.close_connection()

    def open_connection(self):
        self.conn = persistence.Driver().open(self.mbcontext)
        if self.dbname is not None:
            self.conn.set_dbname(self.dbname)

    def begin_transaction(self):
        if self.conn:
            self.conn.transaction_start()
            self.conn.query("BEGIN TRANSACTION")

    def commit(self):
        if self.conn:
            self.conn.query("COMMIT")
            self.conn.transaction_end()

    def rollback(self):
        if self.conn:
            self.conn.query("ROLLBACK")
            self.conn.transaction_end()

    def close_connection(self):
        if self.conn:
            self.conn.close()


class BasePosRepository(object):
    def __init__(self, mbcontext, pos_list):
        # type: (MBEasyContext, list) -> None
        self.mbcontext = mbcontext
        self.pos_list = pos_list
        self.pos_connection_list = []
        self.pos_connection_disct = {}

    def __enter__(self):
        self.open_all_connections()

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close_all_connections()

    def open_all_connections(self):
        for pos_id in self.pos_list:
            pos_connection = persistence.Driver().open(self.mbcontext, dbname=str(pos_id))
            self.pos_connection_list.append(PosConnection(pos_id, pos_connection))
            self.pos_connection_disct[pos_id] = pos_connection

    def close_all_connections(self):
        for pos_connection in self.pos_connection_list:  # type: persistence.Connection
            pos_connection.close()


class BasePosTransactionRepository(BasePosRepository):
    def __init__(self, mbcontext, pos_list):
        super(BasePosTransactionRepository, self).__init__(mbcontext, pos_list)

    def __enter__(self):
        super(BasePosTransactionRepository, self).__enter__()

        self.begin_transaction_on_all_coonections()

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_value:
            self.rollback_on_all_connection()
        else:
            self.commit_on_all_connections()

        super(BasePosTransactionRepository, self).__exit__(exception_type, exception_value, exception_traceback)

    def begin_transaction_on_all_coonections(self):
        for pos_connection in self.pos_connection_list:  # type: PosConnection
            pos_connection.transaction_start()
            pos_connection.query("BEGIN TRANSACTION")

    def commit_on_all_connections(self):
        for pos_connection in self.pos_connection_list:  # type: persistence.Connection
            pos_connection.query("COMMIT")
            pos_connection.transaction_end()

    def rollback_on_all_connection(self):
        for pos_connection in self.pos_connection_list:  # type: persistence.Connection
            pos_connection.query("ROLLBACK")
            pos_connection.transaction_end()



import json
import os
import re

from model.services import ImagesServiceModel
from typing import Dict


class ImagesService(ImagesServiceModel):

    def get_images(
        self,
        images_path,        # type: str
        totem_config_path   # type: str
    ):
        # type: (...) -> Dict

        images = {"navigations": {}, "products": {}, "images": {}, "videos": {}}

        if not images_path or not os.path.exists(images_path):
            return images

        for image_file in os.listdir(images_path):
            full_image_path = os.path.join(images_path, image_file)

            if not os.path.isfile(full_image_path) or not image_file.endswith(".png"):
                continue

            if image_file.startswith("NAV"):
                image_category = "navigations"
            elif image_file.startswith("PRT"):
                image_category = "products"
            else:
                continue

            nav_reference = int(re.findall(r'\d+', image_file)[0])
            image_url = str(os.path.join("sui", "images", image_file)).replace("\\", "/")

            images[image_category][nav_reference] = image_url

        if not totem_config_path or not os.path.exists(totem_config_path):
            return images

        config_file = open(totem_config_path, 'rb')
        data = json.load(config_file)
        images["images"] = data.get("images", {})
        images["videos"] = data.get("videos", {})

        return images



# -*- coding: utf-8 -*-
from datetime import datetime
from typing import List, Dict

import simplejson as json
from application.configurator import Configurator
from application.model import EventData


class TablePictureInteractor(object):

    def __init__(self, configurations):
        # type: (Configurator) -> None
        self.configs = configurations
        self.logger = self.configs.logger

    def build(self, event_data):
        # type: (EventData) -> str

        start_ts = datetime.strptime((event_data.data.get("startTS") + "Z"), "%Y-%m-%d %H:%M:%SZ")
        start_service_date = datetime.strftime(start_ts, "%Y-%m-%dT%H:%M:%SZ")

        finish_ts = event_data.data.get("finishTS")
        finish_ts = finish_ts + 'Z' if finish_ts not in (None, 'None', '') else start_service_date

        linked_tables = event_data.data.get("linkedTables")
        linked_tables = linked_tables.split(",") if linked_tables is not None else []

        payload = {
            "business-dt": event_data.data.get("businessPeriod"),
            "change": float(event_data.data.get("change") or 0),
            "discount-amount": float(event_data.data.get("discountAmount") or 0),
            "due-amount": float(event_data.data.get("dueAmount") or 0),
            "table-id": event_data.data.get("id"),
            "linked-tables": linked_tables,
            "table-seats": int(event_data.data.get("seats") or 1),
            "table-sector": event_data.data.get("sector"),
            "service-id": int(event_data.data.get("serviceId")),
            "service-type": event_data.data.get("type"),
            "service-type-desc": event_data.data.get("typeDescr"),
            "service-seats": int(event_data.data.get("serviceSeats") or 1),
            "slice-source-table-id": event_data.data.get("sliceSourceTableId") or "",
            "slice-source-service-id": self._get_source_service_id(event_data),
            "start-service-dttm": start_service_date,
            "finish-service-dttm": finish_ts,
            "table-status": event_data.data.get("status"),
            "table-status-desc": event_data.data.get("statusDescr"),
            "tax-total": float(event_data.data.get("taxTotal") or 0),
            "tip": float(event_data.data.get("tip") or 0),
            "total-amount": float(event_data.data.get("totalAmount") or 0),
            "total-tender": float(event_data.data.get("totalTender") or 0),
            "user-id": int(event_data.data.get("userId")),
            "service-properties": self._get_custom_properties(event_data),
            "service-orders": self._get_orders(event_data, "ServiceOrders/ServiceOrder"),
            "payment-orders": self._get_orders(event_data, "PaymentOrders/PaymentOrder")
        }

        return json.dumps(payload, use_decimal=True)

    @staticmethod
    def _get_custom_properties(event_data):
        # type: (EventData) -> Dict

        payload_custom_properties = {}
        custom_properties = event_data.data.find("Properties")
        if custom_properties is not None:
            for key, value in custom_properties.items():
                custom_property_key = key.encode("utf-8")
                payload_custom_properties[custom_property_key] = value

        return payload_custom_properties

    def _get_orders(self, event_data, tag):
        # type: (EventData, unicode) -> List

        payload_orders = []
        for order_line in event_data.data.findall("{}".format(tag)):
            order_id = int(order_line.get("orderId") or -1)
            order_state = order_line.get("state")
            created_at = order_line.get("createdAt")
            if created_at is None:
                self.logger.error("Error getting createdAt for order id: {}".format(str(order_id)))
                created_at = datetime.utcfromtimestamp(0).isoformat()

            order_payload = {
                "order-id": order_id,
                "state": order_state,
                "creation-dttm": created_at + "Z"
            }
            payload_orders.append(order_payload)

        return payload_orders

    @staticmethod
    def _get_source_service_id(event_data):
        # type () -> Optional[int]
        if event_data.data.get("sourceServiceId") is not None:
            return int(event_data.data.get("sourceServiceId"))
        return None



class CloseDaySummaryInfo(object):
    def __init__(self, pos_id, store_id, operator_id, business_date, authorizer, open_day_time, initial_date, end_date):
        # type: (int, int, int, unicode, float, unicode, unicode, unicode) -> ()
        self.pos_id = pos_id
        self.store_id = store_id
        self.operator_id = operator_id
        self.business_date = business_date
        self.authorizer = authorizer
        self.open_day_time = open_day_time
        self.initial_date = initial_date
        self.end_date = end_date
